from pendulum.duration import Duration
from rich import box
from rich.console import Console
from rich.table import Table

from timeturner.db import PensiveRow
from timeturner.timeturner import SegmentsByDay

console = Console()


def _pretty_duration(duration: Duration) -> str:
    if duration.seconds < 0:
        return "NEGATIVE TIME"

    periods = [
        ("w", duration.weeks),
        ("d", duration.remaining_days),
        ("h", duration.hours),
        ("m", duration.minutes),
    ]

    parts = []
    for period in periods:
        unit, count = period
        if abs(count) > 0:
            parts.append(f"{count}{unit}")

    return str(" ".join(parts))


def pretty_duration(duration: Duration, breaks: Duration) -> str:
    if breaks:
        return f"{_pretty_duration(duration)} (+{_pretty_duration(breaks)} break)"
    return _pretty_duration(duration)


def segments_by_day(segments: list[SegmentsByDay]) -> None:
    table = Table(
        title="Segments",
        show_header=True,
        header_style="bold magenta",
        show_lines=True,
        box=box.SIMPLE,
    )

    table.add_column("Start", style="dim", width=20)
    table.add_column("End", style="dim", width=20)
    table.add_column("Duration", style="dim", width=20)

    total_durations = []

    for segment in segments:
        total_durations.append(segment.summary.work_time)
        if segment.summary.start:
            start_str = f"{segment.day} {segment.summary.start.format('HH:mm')}"
        else:
            start_str = f"{segment.day}"
        table.add_row(
            start_str,
            segment.summary.end.format("HH:mm") if segment.summary.end else "",
            pretty_duration(segment.summary.work_time, segment.summary.break_time),
        )

    console.print(table)
    total_duration = sum(total_durations, Duration())
    print(f"Total: {total_duration.total_hours():.0f}h {total_duration.minutes}m")


def print_pretty_record(segment: PensiveRow | None) -> None:
    if segment is None:
        console.print("No segment found.")
        return

    print(f"Added record with start: {segment.start.format('YYYY-MM-DD HH:mm')}")



from itertools import groupby
from pathlib import Path
from typing import Iterator, Optional

from pendulum import period
from pendulum.date import Date
from pendulum.duration import Duration
from pendulum.time import Time
from pydantic import BaseModel

from timeturner import loader
from timeturner.db import DatabaseConnection, PensiveRow
from timeturner.parser import parse_args, parse_list_args
from timeturner.tools.boltons_iterutils import pairwise_iter


class DailySummary(BaseModel):
    work_time: Duration
    break_time: Duration
    start: Optional[Time]
    end: Optional[Time]
    by_tag: dict[str, Duration]


class SegmentsByDay(BaseModel):
    day: Date
    weekday: int
    segments: list[PensiveRow]
    summary: DailySummary


def get_summary(segments: list[PensiveRow]) -> DailySummary:
    work_time = Duration()
    break_time = Duration()
    start = None
    end = None
    by_tag = {}
    segment = None
    for segment in segments:
        if start is None:
            start = segment.start.time()
        work_time += segment.duration
        for tag in segment.tags:
            if tag not in by_tag:
                by_tag[tag] = Duration()
            by_tag[tag] += segment.duration

    if segment and segment.end is not None:
        end = segment.end.time()
    for segment, next_segment in pairwise_iter(segments):
        if segment.end is None:
            continue
        new_break = next_segment.start - segment.end
        if new_break >= Duration(minutes=1):
            break_time += new_break

    if work_time > Duration(hours=6, minutes=15):
        if break_time < Duration(minutes=45):
            missing_break_time = Duration(minutes=45) - break_time
            work_time -= missing_break_time
            break_time += missing_break_time
    elif work_time > Duration(hours=4):
        if break_time < Duration(minutes=15):
            missing_break_time = Duration(minutes=15) - break_time
            work_time -= missing_break_time
            break_time += missing_break_time

    return DailySummary(
        work_time=work_time,
        break_time=break_time,
        start=start,
        end=end,
        by_tag=by_tag,
    )


def group_by_day(rows: list[PensiveRow]) -> dict[Date, list[PensiveRow]]:
    return {k: list(v) for k, v in groupby(rows, lambda r: r.start.date())}


def _list(
    time: list[str] | None,
    *,
    db: DatabaseConnection,
) -> list[SegmentsByDay]:
    if time is None:
        time = []
    segments_per_day = {}
    start, end = parse_list_args(time)
    request_period = period(start, end)
    for day in request_period.range("days"):
        segments_per_day[str(day.date())] = []
    rows = db.get_segments_between(start, end)
    for day, segments in groupby(rows, lambda r: r.start.date()):
        segments_per_day[str(day)] = list(segments)
    daily_segments = []
    for day in request_period.range("days"):
        segments = segments_per_day[str(day.date())]
        daily_segments.append(
            SegmentsByDay(
                day=day,
                weekday=day.weekday(),
                segments=segments,
                summary=get_summary(segments),
            )
        )
    return daily_segments


def add(
    time: list[str] | None,
    *,
    db: DatabaseConnection,
) -> PensiveRow:
    if time is None:
        time = []
    start, end = parse_args(time)
    return db.add_segment(start, end)


def end(
    time: list[str] | None,
    *,
    db: DatabaseConnection,
) -> PensiveRow | None:
    if time is None:
        time = []
    end = parse_args(time, single_time=True)
    last_entry = db.get_latest_segment()

    if last_entry is None:
        raise ValueError("No entries to stop")
    if last_entry.end is not None:
        raise ValueError(f"Last entry already ended {last_entry.end!r}")
    pk = last_entry.pk
    db.update_segment(pk, end=end)
    entry = db.get_segment(pk)
    return entry


def import_text(
    path: Path,
    *,
    db: DatabaseConnection,
) -> Iterator[PensiveRow]:
    return loader.import_text(db, path)


def import_json(
    path: Path,
    *,
    db: DatabaseConnection,
) -> Iterator[PensiveRow]:
    return loader.import_json(db, path)



import re
from pathlib import Path
from typing import Iterator, cast

from pendulum.datetime import DateTime
from pendulum.parser import parse

from timeturner.db import DatabaseConnection, PensiveRow, TimeSegment

re_date = re.compile(r"^\d{4}-\d{2}-\d{2}")


def extract_segments(lines: list[str]) -> Iterator[TimeSegment]:
    segment = None
    for line in lines:
        line = line.strip()

        if line[0].isdigit():
            # this is a new segment
            if segment is not None:
                yield segment
        elif line[0] == "#":
            # this is a continuation of the previous segment
            if segment is None:  # pragma: no cover
                raise ValueError("Continuation without segment")
            segment.tags = [e[1:].strip() for e in line.split(",")]
            continue
        else:
            # this is a continuation of the previous segment
            if segment is None:  # pragma: no cover
                raise ValueError("Continuation without segment")
            if segment.description is None:  # pragma: no cover
                segment.description = line
            else:
                segment.description += f"\n{line}"
            continue

        # split the line into columns
        columns = line.split("|")

        # remove the whitespace
        columns = [c.strip() for c in columns]

        # parse the start and end date
        start = cast(DateTime, parse(columns[0]))
        end = cast(DateTime, parse(columns[1])) if columns[1] else None

        # duration is not used
        # columns[2]

        # parse the activity
        activity = columns[3]

        # parse the tags
        category = columns[4]

        # build the TimeSegment object

        segment = TimeSegment(
            start=start,
            end=end,
            passive=category == "travel",
            tags=[],
            description=f"{category}@{activity}",
        )

    if segment is not None:
        yield segment


def import_text(db: DatabaseConnection, text_file: Path) -> Iterator[PensiveRow]:
    table_separator = "-" * 80 + "\n"

    with text_file.open() as f:
        lines = f.readlines()

    # find the table separator
    try:
        index = lines.index(table_separator)
    except ValueError:
        raise ValueError(f"Could not find table separator {table_separator}")

    # remove the header
    lines = lines[index + 1 :]

    # remove the footer
    lines = lines[: lines.index(table_separator)]

    for entry in extract_segments(lines):
        yield db.add_segment(**entry.dict())


def import_json(db: DatabaseConnection, json_file: Path) -> Iterator[PensiveRow]:
    import json

    with json_file.open() as f:
        data = json.load(f)

    for entry in data:
        yield db.add_segment(
            start=entry["start"],
            end=entry["end"],
            passive=entry["passive"],
            tags=entry["tags"],
            description=entry["description"],
        )



from enum import Enum
from pathlib import Path
from typing import Literal, Optional

import typer
from pydantic.json import pydantic_encoder
from rich.console import Console

from timeturner import rich_output, timeturner
from timeturner.db import DatabaseConnection
from timeturner.settings import TimeTurnerSettings

app = typer.Typer()

console = Console()
print = console.print
settings = TimeTurnerSettings()


class Output(str, Enum):
    json = "json"
    rich = "rich"


@app.callback()
def callback(
    output: Optional[Output] = None,
):
    global settings
    additional_settings_args = dict()
    if output is not None:
        additional_settings_args["output"] = output
    if additional_settings_args:
        settings = TimeTurnerSettings(**additional_settings_args)


@app.command("l", hidden=True)
@app.command("list")
def _list(time: Optional[list[str]] = typer.Argument(None)):
    data = timeturner._list(time, db=settings.database.connection)
    if settings.output == "json":
        console.print_json(data=data, default=pydantic_encoder)
    else:
        rich_output.segments_by_day(data)


@app.command("a", hidden=True)
@app.command()
def add(
    time: Optional[list[str]] = typer.Argument(None),
    auto_stop: Optional[bool] = None,
    # passive: bool | None = False,
    # tags: str | None = None,
    # description: str | None = None,
):
    data = timeturner.add(time, db=settings.database.connection)
    if settings.output == "json":
        console.print_json(data=data, default=pydantic_encoder)
    else:
        rich_output.print_pretty_record(data)


@app.command("e", hidden=True)
@app.command("end")
def end(
    time: Optional[list[str]] = typer.Argument(None),
):
    data = timeturner.end(time, db=settings.database.connection)
    console.print_json(data=data, default=pydantic_encoder)


@app.command("i", hidden=True)
@app.command(name="import")
def import_(text_file: Path):
    data = list(timeturner.import_json(text_file, db=settings.database.connection))
    console.print_json(data=data, default=pydantic_encoder)


@app.command("c", hidden=True)
@app.command(name="config")
def config():
    console.print_json(data=settings, default=pydantic_encoder)


def entrypoint():
    app()


if __name__ == "__main__":
    entrypoint()



import re

# from pendulum.tz.timezone import Timezone
from enum import Enum
from typing import Any, Literal, TypedDict, overload

import pendulum
from pendulum.datetime import DateTime
from pendulum.duration import Duration

delta_components = re.compile(r"(?P<sign>[+-])?(?P<value>\d+)(?P<unit>[mhd])")
range_components = re.compile(r"(?P<value>\d+)?(?P<unit>[dwMyY]|day|week|month|year)s?")


class ComponentType(Enum):
    TIME = "time"
    DATE = "date"
    DELTA = "delta"
    DELTA_WITH_TIME = "delta_with_time"
    ALIAS = "alias"
    RANGE = "range"
    NUMBER = "number"
    UNKNOWN = "unknown"


class ListGroupBy(Enum):
    AUTO = "auto"
    DONT_GROUP = "dont_group"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    YEAR = "year"


class DateDict(TypedDict, total=False):
    year: int
    month: int
    day: int


class TimeDict(TypedDict, total=False):
    hour: int
    minute: int
    second: int


class DateTimeDict(TypedDict, total=False):
    year: int
    month: int
    day: int
    hour: int
    minute: int
    second: int
    tzinfo: Any


def split_array(array: list[str], separator: str) -> tuple[list[str], list[str]]:
    try:
        index = array.index(separator)
    except ValueError:
        return array, []
    return array[:index], array[index + 1 :]


def get_component_type(component: str, *, include_list_args) -> ComponentType:
    if include_list_args:
        if component in ["today", "yesterday", "week", "month", "year"]:
            return ComponentType.ALIAS
        if range_components.match(component):
            return ComponentType.RANGE
    if component.startswith("-") or component.startswith("+"):
        if "@" in component:
            return ComponentType.DELTA_WITH_TIME
        return ComponentType.DELTA
    if ":" in component:
        return ComponentType.TIME
    if "-" in component:
        return ComponentType.DATE
    if component.isnumeric():
        return ComponentType.NUMBER
    return ComponentType.UNKNOWN


def parse_time(time: str, now: DateTime) -> DateTime:
    ret = TimeDict()
    match time.split(":"):
        case [hour, minute]:
            ret["hour"] = int(hour)
            ret["minute"] = int(minute)
        case [hour, minute, second]:
            ret["hour"] = int(hour)
            ret["minute"] = int(minute)
            ret["second"] = int(second)
        case _:
            raise ValueError(f"Invalid time: {time}")
    return now.set(**ret)


def parse_date(date: str, now: DateTime) -> DateTime:
    ret = DateDict()
    match date.split("-"):
        case [year, month, day]:
            ret["year"] = int(year)
            ret["month"] = int(month)
            ret["day"] = int(day)
        case [month, day]:
            ret["month"] = int(month)
            ret["day"] = int(day)
        case [day]:
            ret["day"] = int(day)
        case _:
            raise ValueError(f"Invalid date: {date}")
    return now.set(**ret)


def parse_delta(delta: str, now: DateTime) -> DateTime:
    ret = dict()
    matches = list(delta_components.finditer(delta))
    if not matches:
        raise ValueError(f"Invalid delta: {delta}")
    for match in matches:
        sign = match.group("sign")
        value = int(match.group("value"))
        unit = match.group("unit")
        if sign == "-":
            value = -value
        if unit == "d":
            ret["days"] = value
        elif unit == "h":
            ret["hours"] = value
        elif unit == "m":  # pragma: no branch
            ret["minutes"] = value

    return now + Duration(**ret)


def parse_delta_with_time(delta_with_time: str, now: DateTime) -> DateTime:
    DateTimeDict()
    delta, time = delta_with_time.split("@")
    now = parse_delta(delta, now)
    return parse_time(time, now)


def single_time_parse(
    components: list[str],
    *,
    now: DateTime | None = None,
    include_list_args: bool = False,
) -> DateTime:
    if now is None:
        now = pendulum.now()

    components_with_types = [
        (component, get_component_type(component, include_list_args=include_list_args))
        for component in components
    ]
    # print(components_with_types)
    match components_with_types:
        case [(time, ComponentType.TIME)]:
            now = parse_time(time, now)
        case [(date, ComponentType.DATE)] | [(date, ComponentType.NUMBER)]:
            now = parse_date(date, now)
        case [(delta, ComponentType.DELTA)]:
            now = parse_delta(delta, now)
        case [(delta, ComponentType.DELTA), (time, ComponentType.TIME)]:
            now = parse_delta(delta, now)
            now = parse_time(time, now)
        case [(delta_with_time, ComponentType.DELTA_WITH_TIME)]:
            now = parse_delta_with_time(delta_with_time, now)
        case [(date, ComponentType.DATE), (time, ComponentType.TIME)] | [
            (date, ComponentType.NUMBER),
            (time, ComponentType.TIME),
        ]:
            now = parse_date(date, now)
            now = parse_time(time, now)
        case []:
            pass
        case _:
            raise ValueError(f"Invalid components: {components}")
    return now.set(second=0, microsecond=0)


@overload
def parse_args(
    args: list[str],
    *,
    single_time: Literal[True],
    prefer_full_days: bool = False,
) -> DateTime:
    ...


@overload
def parse_args(
    args: list[str],
    *,
    prefer_full_days: Literal[True],
) -> tuple[DateTime, DateTime]:
    ...


@overload
def parse_args(
    args: list[str],
    *,
    prefer_full_days: Literal[False],
) -> tuple[DateTime, DateTime | None]:
    ...


@overload
def parse_args(
    args: list[str],
) -> tuple[DateTime, DateTime | None]:
    ...


def parse_args(
    args: list[str],
    *,
    prefer_full_days: bool = False,
    single_time: bool = False,
) -> tuple[DateTime, DateTime | None] | DateTime:
    start, end = split_array(args, "-")
    start = single_time_parse(start)
    if end:
        end = single_time_parse(end, now=start)
    else:
        end = None
    if prefer_full_days:
        start = start.start_of("day")
        if not end:
            end = start.end_of("day")
        else:
            end = end.end_of("day")
    if single_time:
        return start
    return start, end


def parse_list_args(
    args: list[str],
    *,
    now: DateTime | None = None,
) -> tuple[DateTime, DateTime]:
    if now is None:
        now = pendulum.now()
    now = now.set(second=0, microsecond=0)
    start = now.start_of("day")
    end = now.end_of("day")
    if "-" in args:
        start, end = split_array(args, "-")
        start = single_time_parse(start, now=now)
        end = single_time_parse(end, now=start)

        if start.time() == now.time():
            start = start.start_of("day")
        if end.time() == now.time():
            end = end.end_of("day")

    elif len(args) == 0:
        pass
    elif len(args) == 1:
        arg = args[0]
        if arg in ["week", "month", "year"]:
            start = now.start_of(arg)
            end = now.end_of(arg)
        elif arg == "today":
            pass
        elif arg == "yesterday":
            start = start.subtract(days=1)
            end = end.subtract(days=1)
        elif _match := range_components.match(arg):
            match _match.groupdict():
                case {"value": value, "unit": "d" | "day" | "days"}:
                    start = start.subtract(days=int(value) - 1)
                case {"value": value, "unit": "w" | "week" | "weeks"}:
                    start = start.start_of("week").subtract(weeks=int(value) - 1)
                    end = now.end_of("week")
                case {"value": value, "unit": "M" | "month" | "months"}:
                    start = start.start_of("month").subtract(months=int(value) - 1)
                    end = now.end_of("month")
                case {  # pragma: no branch
                    "value": value,
                    "unit": "y" | "year" | "years",
                }:
                    start = start.start_of("year").subtract(years=int(value) - 1)
                    end = now.end_of("year")

        else:
            start = single_time_parse(args, now=now).start_of("day")
    else:
        # we have 2 arguments, but no dash
        # we definitely define a specific time, so we don't change to
        # the start of the day, or the end of the day
        start = single_time_parse(args, now=now)
        print(start.time())
        print(now.time())

    return start, end



"""
Here we define the database models for the time tracking application.

The database is a SQLite database, we store all time segments in the pensieve table.

It has the following columns:

    pk: primary key, autoincrementing integer
    start: start time of the time segment, in ISO 8601 format
    end: end time of the time segment, in ISO 8601 format
    passive: whether the time segment was passive or not, boolean, default False
    tags: tags associated with the time segment, comma separated string
    description: description of the time segment, string

"""

import sqlite3
from datetime import datetime
from typing import Any, cast

import pendulum
from pendulum.datetime import DateTime
from pendulum.parser import parse
from pendulum.period import Period
from pydantic import BaseModel, validator


class Sentinel:
    ...


sentinel = Sentinel()


class TimeSegment(BaseModel):
    start: DateTime
    end: DateTime | None = None
    passive: bool = False
    tags: list[str] = []
    description: str | None = None

    @validator("start")
    def parse_start(cls, value: str | datetime | DateTime) -> DateTime:
        if isinstance(value, DateTime):
            return value
        if isinstance(value, datetime):
            value = str(value)
        new_value = parse(value)
        if isinstance(new_value, DateTime):
            return new_value
        raise ValueError(f"Could not parse {value} as a datetime")

    @validator("end")
    def parse_end(cls, value: str | datetime | DateTime | None) -> DateTime | None:
        if value is None:
            return None
        return cls.parse_start(value)

    @validator("tags")
    def parse_tags(cls, value: str | list[str] | None) -> list[str] | None:
        if not value:
            # we want to return a new empty list, not reuse the default empty list
            return []
        if isinstance(value, str):
            return value.split(",")
        if isinstance(value, list):
            return value
        raise ValueError(f"Could not parse {value} as a list of tags")

    @property
    def duration(self) -> Period:
        if self.end is None:
            duration = cast(Period, pendulum.now() - self.start)
        else:
            duration = cast(Period, self.end - self.start)

        return duration


class PensiveRow(TimeSegment):
    pk: int


def str_if_not_none(value: Any) -> str | None:
    if value is None:
        return None
    return str(value)


class DatabaseConnection:
    def __init__(
        self,
        database_file: str = "timeturner.db",
        table_name: str = "pensieve",
    ):
        self.table_name = table_name
        self.database_file = database_file
        self.connection = sqlite3.connect(self.database_file)
        self.create_table()

    def create_table(self):
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self.table_name} (
                pk INTEGER PRIMARY KEY AUTOINCREMENT,
                start DATETIME,
                end DATETIME,
                passive BOOLEAN DEFAULT FALSE,
                description TEXT
            )
            """
        )
        cursor.execute(
            f"""
            CREATE TABLE IF NOT EXISTS tags (
                pk INTEGER PRIMARY KEY AUTOINCREMENT,
                tag TEXT
            )
            """
        )
        cursor.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self.table_name}_tags (
                {self.table_name}_pk INTEGER,
                tags_pk INTEGER,
                FOREIGN KEY({self.table_name}_pk) REFERENCES {self.table_name}(pk),
                FOREIGN KEY(tags_pk) REFERENCES tags(pk)
            )
            """
        )
        self.connection.commit()

    def add_segment(
        self,
        start: DateTime,
        end: DateTime | None = None,
        passive: bool | None = False,
        tags: list[str] | None = None,
        description: str | None = None,
    ) -> PensiveRow:
        # fix passive if it is None
        if passive is None:
            passive = False
        if tags is None:
            tags = []
        query = f"""
            INSERT INTO {self.table_name} (start, end, passive, description)
            VALUES (?, ?, ?, ?)
            """
        values = (
            str_if_not_none(start),
            str_if_not_none(end),
            passive,
            description,
        )
        cursor = self.connection.cursor()
        cursor.execute(
            query,
            values,
        )
        pk = cursor.lastrowid
        for tag in tags:
            tag_pk = self.insert_or_get_tag_pk(tag)
            cursor.execute(
                f"""
                INSERT INTO {self.table_name}_tags ({self.table_name}_pk, tags_pk)
                VALUES (?, ?)
                """,
                (pk, tag_pk),
            )
        self.connection.commit()
        if cursor.lastrowid is None:
            raise Exception("Failed to insert segment into database")
        return cast(PensiveRow, self.get_segment(pk))

    def insert_or_get_tag_pk(self, tag: str) -> int:
        cursor = self.connection.cursor()
        cursor.execute(
            """
            SELECT pk FROM tags WHERE tag = ?
            """,
            (tag,),
        )
        result = cursor.fetchone()
        if result is None:
            cursor.execute(
                """
                INSERT INTO tags (tag) VALUES (?)
                """,
                (tag,),
            )
            self.connection.commit()
            return cursor.lastrowid
        return result[0]

    def get_tags_for_segment(self, pk: int) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
            SELECT tags.tag FROM {self.table_name}_tags
            INNER JOIN tags ON tags.pk = {self.table_name}_tags.tags_pk
            WHERE {self.table_name}_tags.{self.table_name}_pk = ?
            """,
            (pk,),
        )
        return [tag for tag, in cursor.fetchall()]

    def update_segment(
        self,
        pk: int,
        start: DateTime | Sentinel = sentinel,
        end: DateTime | None | Sentinel = sentinel,
        passive: bool | None | Sentinel = sentinel,
        tags: list[str] | None | Sentinel = sentinel,
        description: str | None | Sentinel = sentinel,
    ) -> None:
        elements_to_update = []
        values_to_update = []
        if start is not sentinel:
            elements_to_update.append("start = ?")
            values_to_update.append(str(start))
        if end is not sentinel:
            elements_to_update.append("end = ?")
            values_to_update.append(str_if_not_none(end))
        if passive is not sentinel:
            elements_to_update.append("passive = ?")
            if passive is None:
                passive = False
            values_to_update.append(passive)
        if description is not sentinel:
            elements_to_update.append("description = ?")
            values_to_update.append(description)

        query = f"""
            UPDATE {self.table_name}
            SET {', '.join(elements_to_update)}
            WHERE pk = ?
            """
        values = (
            *values_to_update,
            pk,
        )
        cursor = self.connection.cursor()
        cursor.execute(
            query,
            values,
        )

        # insert all tags that are not already in the database
        if tags is not sentinel:
            if tags is None:
                tags = []
            stored_tags = set(self.get_tags_for_segment(pk))
            tags_to_insert = set(tags) - stored_tags
            for tag in tags_to_insert:
                tag_pk = self.insert_or_get_tag_pk(tag)
                cursor.execute(
                    f"""
                    INSERT INTO {self.table_name}_tags ({self.table_name}_pk, tags_pk)
                    VALUES (?, ?)
                    """,
                    (pk, tag_pk),
                )
            tags_to_remove = stored_tags - set(tags)
            for tag in tags_to_remove:
                cursor.execute(
                    f"""
                    DELETE FROM {self.table_name}_tags
                    WHERE {self.table_name}_pk = ? AND tags_pk = ?
                    """,
                    (pk, self.insert_or_get_tag_pk(tag)),
                )

                # if tag is not used anywhere else, delete it from the tags table
                cursor.execute(
                    f"""
                    SELECT COUNT(*) FROM tags
                    INNER JOIN {self.table_name}_tags ON tags.pk = {self.table_name}_tags.tags_pk
                    WHERE tags.tag = ?
                    """,
                    (tag,),
                )
                count = cursor.fetchone()
                if count is None:
                    raise Exception("Failed to get count of tags")
                if count[0] == 0:
                    cursor.execute(
                        """
                        DELETE FROM tags WHERE tag = ?
                        """,
                        (tag,),
                    )

        self.connection.commit()

    def get_latest_segment(self) -> PensiveRow | None:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
            SELECT pk, start, end, passive, description FROM {self.table_name}
            ORDER BY start DESC LIMIT 1
            """
        )

        row = cursor.fetchone()
        if row is None:
            return None
        pk, start, end, passive, description = row
        if passive is None:
            passive = False
        return PensiveRow(
            pk=pk,
            start=start,
            end=end,
            passive=passive,
            tags=self.get_tags_for_segment(pk),
            description=description,
        )

    def delete_segment(self, pk: int) -> None:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
            DELETE FROM {self.table_name} WHERE pk = ?
            """,
            (pk,),
        )
        self.connection.commit()

    def get_segment(self, pk: int) -> PensiveRow | None:
        cursor = self.connection.cursor()
        cursor.execute(
            f"""
            SELECT pk, start, end, passive, description FROM {self.table_name}
            WHERE pk = ?
            """,
            (pk,),
        )

        row = cursor.fetchone()
        if row is None:
            return None
        pk, start, end, passive, description = row
        if passive is None:
            passive = False
        return PensiveRow(
            pk=pk,
            start=start,
            end=end,
            passive=passive,
            tags=self.get_tags_for_segment(pk),
            description=description,
        )

    def get_segments_between(
        self,
        start: DateTime,
        end: DateTime,
    ) -> list[PensiveRow]:
        """
        Get all segments between the given start and end times.

        This includes segments that start before the given start time and / or end after
        the given end time.
        """
        cursor = self.connection.cursor()

        cursor.execute(
            f"""
            SELECT pk, start, end, passive, description FROM {self.table_name}
            WHERE start <= ? AND (end >= ? OR end IS NULL)
            ORDER BY start ASC
            """,
            (str(end), str(start)),
        )

        rows = cursor.fetchall()
        return [
            PensiveRow(
                pk=pk,
                start=start,
                end=end,
                passive=passive,
                tags=self.get_tags_for_segment(pk),
                description=description,
            )
            for pk, start, end, passive, description in rows
        ]

    def get_all_segments(self) -> list[PensiveRow]:
        cursor = self.connection.cursor()

        cursor.execute(
            f"""
            SELECT pk, start, end, passive, description FROM {self.table_name}
            """
        )

        rows = cursor.fetchall()
        return [
            PensiveRow(
                pk=pk,
                start=start,
                end=end,
                passive=passive,
                tags=self.get_tags_for_segment(pk),
                description=description,
            )
            for pk, start, end, passive, description in rows
        ]



from os import environ
from pathlib import Path
from typing import Any, Literal

import tomlkit
from pydantic import BaseModel, BaseSettings

from timeturner.db import DatabaseConnection

DEFAULT_CONFIG_HOME = (
    Path.home() / environ.get("XDG_CONFIG_HOME", ".config") / "timeturner"
)


class Settings(BaseSettings):
    config_file: str = "config.toml"
    config_home: Path = DEFAULT_CONFIG_HOME

    @property
    def config_path(self) -> Path:
        return self.config_home / self.config_file

    class Config:
        env_file_encoding = "utf-8"
        env_prefix = "timeturner_"


config_settings = Settings()


def load_config_file(settings: BaseSettings) -> dict[str, Any]:
    encoding = settings.__config__.env_file_encoding
    config_file = config_settings.config_path
    if not config_file.exists():
        return {}
    return tomlkit.loads(config_file.read_text(encoding))


class DatabaseSettings(BaseModel):
    file: str = "timeturner.db"
    home: Path = config_settings.config_home
    table_name: str = "pensieve"

    @property
    def database_path(self) -> Path:
        return self.home / self.file

    @property
    def connection(self) -> DatabaseConnection:
        self.home.mkdir(parents=True, exist_ok=True)
        return DatabaseConnection(
            str(self.database_path.resolve()),
            table_name=self.table_name,
        )


class TimeTurnerSettings(Settings):
    database: DatabaseSettings = DatabaseSettings()
    output: Literal["json", "rich"] = "rich"

    class Config:
        env_file_encoding = "utf-8"
        env_prefix = "timeturner_"
        env_nested_delimiter = "__"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            del file_secret_settings
            return (
                init_settings,
                env_settings,
                load_config_file,
            )



# taken from: https://github.com/mahmoud/boltons/blob/master/boltons/iterutils.py
# extracted needed functions and removed python 2 support

# Copyright (c) 2013, Mahmoud Hashemi
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * The names of the contributors may not be used to endorse or
#      promote products derived from this software without specific
#      prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import itertools
from typing import Iterable, Iterator, TypeVar

T = TypeVar("T")


def windowed_iter(src: Iterable[T], size: int) -> Iterator[tuple[T, ...]]:
    """Returns tuples with length *size* which represent a sliding
    window over iterable *src*.
    >>> list(windowed_iter(range(7), 3))
    [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5), (4, 5, 6)]
    If the iterable is too short to make a window of length *size*,
    then no window tuples are returned.
    >>> list(windowed_iter(range(3), 5))
    []
    """
    tees = itertools.tee(src, size)
    try:
        for i, t in enumerate(tees):
            for _ in range(i):
                next(t)
    except StopIteration:
        return zip([])
    return zip(*tees)


def pairwise_iter(src: Iterable[T]) -> Iterator[tuple[T, T]]:
    """Returns tuples of adjacent items from iterable *src*.
    >>> list(pairwise_iter(range(5)))
    [(0, 1), (1, 2), (2, 3), (3, 4)]
    If the iterable is too short to make a window of length 2,
    then no window tuples are returned.
    >>> list(pairwise_iter(range(1)))
    []
    """
    return windowed_iter(src, 2)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['timeturner', 'timeturner.tools']

package_data = \
{'': ['*']}

install_requires = \
['pendulum>=2.1.2,<3.0.0',
 'pydantic>=1.10.5,<2.0.0',
 'rich>=13.3.1,<14.0.0',
 'tomlkit>=0.11.6,<0.12.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['timeturner = timeturner.run:entrypoint',
                     'tt = timeturner.run:entrypoint']}

setup_kwargs = {
    'name': 'timeturner',
    'version': '0.1.1',
    'description': '',
    'long_description': 'Time Turner\n===========\n\n## A Minimalistic Time Tracker.\n\nThis is a minimalistic time tracker that allows you to record when you start working, even if it is in the past, stop running activities now, and add activities from the past. It also ensures legal rest periods are included if you forgot to track them.\n\nIn the Harry Potter series, the Time-Turner is a magical device that allows the user to travel back in time. Time Turner is a time tracker that lets you "turn back time" and record activities from the past.\n\n**Warning, this is still an alpha release, the API is not stable yet.**\n\n## Usage\n\n### Starting an Activity\n\nTo start tracking an activity, run the following command:\n\n\n![`timeturner add`](img/add.svg)\n\nThis will record the current time as the start of your activity.\n\nIf you forgot to start tracking an activity yesterday, you can provide the start time with an additional parameter, `-10m` would mean 10 minutest ago. The full list of possible time values can be [seen further down](#examples-for-times)\n\n### Stopping an Activity\n\nTo stop tracking the current activity, run the following command:\n\n\n![`timeturner end`](img/end.svg)\n\n\nThis will record the current time as the end of your activity and calculate the total duration.\n\n### Adding a Past Activity\n\nIf you forgot to track an activity in the past, you can add it with `timeturner add <start_time> - <end_time>`\n\n![`timeturner add -- -1d@9:00 - +8h45m`](img/add_past.svg)\n\n## Configuration\n\n| Environment Variable       | Default Value                    | Description                                  |\n| -------------------------- | -------------------------------- | -------------------------------------------- |\n| TIMETURNER_CONFIG_HOME     | ~/$XDG_CONFIG_HOME/timeturner    | The directory for configuration files.       |\n| TIMETURNER_CONFIG_FILE     | timeturner.toml                  | The configuration file to use.               |\n| TIMETURNER_DATABASE__HOME  | value of $TIMETURNER_CONFIG_HOME | The directory to store the database file in. |\n| TIMETURNER_DATABASE__FILE  | timeturner.db                    | The database file to use.                    |\n| TIMETURNER_DATABASE__TABLE | pensive                          | The table to use in the database.            |\n\n## Examples\n\n### Examples for times\n\n<start_time> or <end_time> could be one of the following Examples:\n\n| Example         | Description                               |\n| --------------- | ----------------------------------------- |\n|                 | now                                       |\n| 9:00            | 9:00 today                                |\n| -1m             | 1 minute ago                              |\n| -1h             | 1 hour ago                                |\n| -1d             | 1 day ago, you will be asked for the time |\n| -1d@9:00        | 1 day ago 9:00                            |\n| +1m             | 1 minute from now                         |\n| +1h             | 1 hour from now                           |\n| 12 7:00         | 7:00 on the 12th of the current month     |\n| 02-28 9:00      | 9:00 on February 28 of the current year   |\n| 2022-02-28 9:00 | 9:00 on February 28, 2022                 |\n\n\n\n### Automatic Rest Periods\n\nIf you forget to track a rest period, the time tracker will reduce the required rest period before showing it. For periods greater than 4h 15 are reducted, for periods greater than 6:15 additional 30m are reducted.\n\n\n\nTODOs:\n- [ ] Add Changelog\n- [ ] Add Configuration\n  - [ ] ignore seconds\n  - [ ] freeze time, to generate useful images\n  - [ ] automatic rest periods\n- [ ] show and generate tui output\n- [ ] Add section about contributions\n- [ ] Add precommit hook to ensure code is formatted\n- [ ] Add version and git hash to build, so it can be shown in the tui\n- [ ] Generate docstrings for DB methods\n- [ ] Remove import command (it contains assumptions that will not be true for everyone)\n  - [ ] Document how to import data from other time trackers\n  - [ ] Add mode to convert hamster output to jsonl file.\n  - [ ] Add mode to import jsonl file\n- [ ] Add logging\n  - [ ] allow different log levels for database and application\n- [ ] add a change log\n- [ ] README\n  - [ ] auto generate config options\n- [ ] db\n  - [ ] add table for tags\n\n\nTODOS by command:\n\n- [ ] add\n  - [ ] auto close activities that are still open\n  - [ ] break up activities when a activity is inserted in the middle\n  - [ ] allow passive activities\n  - [ ] allow sick days\n  - [ ] add tests\n- [ ] end\n  - [ ] add tests\n\n- [ ] configure\n  - [ ] modify and write configfile\n  - [ ] allow to add aliases for commands\n    - [ ] e.g. new new add alias with setting passive to true\n- [ ] list\n  - [ ] split up multiday activities\n  - [ ] add option to show only open activities\n  - [ ] group by year, month, week, day\n  - [ ] add tests\n\n### Design Goals\n\n- minimalistic, little to type\n- enforce as little as possible\n- be clear\n- be extensible\n  - TODO: support plugins (maybe a later version)\n  - be able to use it programmatically\n  - be able to use it as a library\n\n\n### Open Questions\n\n- [ ] should the get_latest_segment return segments from the future (start_time in the future)?\n\n### Nice to have:\n- [ ] Build a minimal Docker image (maybe)\n- [ ] https://github.com/ines/termynal\n',
    'author': 'Olaf Gladis',
    'author_email': 'olaf@gladis.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
